#
# Copyright (C) 2012 INRA
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import argparse
import time

try:
    import _preamble
except ImportError:
    sys.exc_clear()

from jflow.workflows_manager import WorkflowsManager
from jflow.workflow import Workflow
import jflow.utils as utils
from cctools.util import time_format

class JflowArgumentParser (argparse.ArgumentParser):
    def _read_args_from_files(self, arg_strings):
        # expand arguments referencing files
        new_arg_strings = []
        for arg_string in arg_strings:
            # if it's not a comment or an empty line
            if not arg_string.startswith("#") and arg_string:
                # for regular arguments, just add them back into the list
                if not arg_string or arg_string[0] not in self.fromfile_prefix_chars:
                    new_arg_strings.append(arg_string)
                # replace arguments referencing files with the file content
                else:
                    try:
                        with open(arg_string[1:]) as args_file:
                            arg_strings = []
                            # give to the convert_arg_line_to_args a table of lines instead of line per line
                            for arg in self.convert_arg_line_to_args(args_file.read().splitlines()):
                                arg_strings.append(arg)
                            arg_strings = self._read_args_from_files(arg_strings)
                            new_arg_strings.extend(arg_strings)
                    except OSError:
                        err = _sys.exc_info()[1]
                        self.error(str(err))
        # return the modified argument list
        return new_arg_strings

def display_workflow_status(workflow, detailed=False):
    if workflow.start_time: start_time = time.asctime(time.localtime(workflow.start_time))
    else: start_time = "-"
    if workflow.start_time and workflow.end_time: elapsed_time = str(workflow.end_time-workflow.start_time)
    elif workflow.start_time: elapsed_time = str(time.time()-workflow.start_time)
    else: elapsed_time = "-"
    if workflow.end_time: end_time = time.asctime(time.localtime(workflow.end_time))
    else: end_time = "-"
    if detailed:
        status = "Workflow #" + utils.get_nb_string(workflow.id) + " (" + workflow.name + ") is " + \
                workflow.get_status() + ", time elapsed: " + str(elapsed_time) + " (from " + start_time + \
                " to " + end_time + "):\n"
        for i, component in enumerate(workflow.get_components_nameid()):
            status_info = workflow.get_component_status(component)
            try: perc_waiting = (status_info["waiting"]*100.0)/status_info["tasks"]
            except: perc_waiting = 0
            try: perc_running = (status_info["running"]*100.0)/status_info["tasks"]
            except: perc_running = 0
            try: perc_failed = (status_info["failed"]*100.0)/status_info["tasks"]
            except: perc_failed = 0
            try: perc_aborted = (status_info["aborted"]*100.0)/status_info["tasks"]
            except: perc_aborted = 0
            try: perc_completed = (status_info["completed"]*100.0)/status_info["tasks"]
            except: perc_completed = 0
            
            if status_info["running"] > 0: running = "\033[94mrunning:" + str(status_info["running"]) + "\033[0m"
            else: running = "running:" + str(status_info["running"])
            if status_info["waiting"] > 0: waiting = "\033[93mwaiting:" + str(status_info["waiting"]) + "\033[0m"
            else: waiting = "waiting:" + str(status_info["waiting"])            
            if status_info["failed"] > 0: failed = "\033[91mfailed:" + str(status_info["failed"]) + "\033[0m"
            else: failed = "failed:" + str(status_info["failed"])
            if status_info["aborted"] > 0: aborted = "\033[95maborted:" + str(status_info["aborted"]) + "\033[0m"
            else: aborted = "aborted:" + str(status_info["aborted"])
            if status_info["completed"] == status_info["tasks"] and status_info["completed"] > 0: completed = "\033[92mcompleted:" + str(status_info["completed"]) + "\033[0m"
            else: completed = "completed:" + str(status_info["completed"])
                        
            status += "  - " + component + ", time elapsed " + time_format(status_info["time"]) + \
                " (total:" + str(status_info["tasks"]) + ", " + waiting + ", " + running + ", " + failed + \
                ", " + aborted + ", " + completed + ")"
            if i<len(workflow.get_components_nameid())-1: status += "\n"
        return status
    else:
        return utils.get_nb_string(workflow.id) + "\t" + workflow.name + "\t" + workflow.get_status() + \
            "\t" + elapsed_time + "\t" + start_time + "\t" + end_time

if __name__ == '__main__':

    # Create a workflow manager to get access to our workflows
    wfmanager = WorkflowsManager()
    
    # Create the top-level parser
    parser = JflowArgumentParser()
    subparsers = parser.add_subparsers(title='Available sub commands')
    
    # Add rerun workflow availability
    sub_parser = subparsers.add_parser("rerun", help="rerun a specific workflow")
    sub_parser.add_argument("--workflow-id", type=str, help="Which workflow should be rerun",
                            required=True, dest="workflow_id")
    sub_parser.set_defaults(cmd_object="rerun")

    # Add status workflow availability
    sub_parser = subparsers.add_parser("status", help="monitor a specific workflow")
    sub_parser.add_argument("--workflow-id", type=str, help="Which workflow status should be displayed",
                            default=None, dest="workflow_id")
    sub_parser.add_argument("--all", action="store_true", help="Display all workflows status",
                            default=False, dest="all")
    sub_parser.set_defaults(cmd_object="status")
    
    # Add available pipelines
    wf_instances = wfmanager.get_available_workflows()
    wf_classes = []
    for instance in wf_instances:
        wf_classes.append(instance.__class__.__name__)
        # create the subparser for each applications
        sub_parser = subparsers.add_parser(instance.name, help=instance.description, fromfile_prefix_chars='@')
        sub_parser.convert_arg_line_to_args = instance.__class__.config_parser
        [parameters_groups, parameters_order] = instance.get_parameters_per_groups()
        for group in parameters_order:            
            if group == "default":
                for param in parameters_groups[group]:
                    sub_parser.add_argument(param.flag, **param.export_to_argparse())
            elif group.startswith("exclude-"):
                is_required = False
                for param in parameters_groups[group]:
                    if param.required:
                        is_required = True
                        # an exlcusive parameter cannot be required, the require is at the group level
                        param.required = False
                pgroup = sub_parser.add_mutually_exclusive_group(required=is_required)
                for param in parameters_groups[group]:
                    pgroup.add_argument(param.flag, **param.export_to_argparse())
            else:
                pgroup = sub_parser.add_argument_group(group)
                for param in parameters_groups[group]:
                    pgroup.add_argument(param.flag, **param.export_to_argparse())
        sub_parser.set_defaults(cmd_object=instance.__class__.__name__)
    args = vars(parser.parse_args())
        
    if args["cmd_object"] in wf_classes:
        wfmanager.run_workflow(args["cmd_object"], args)
    elif args["cmd_object"] == "rerun":
        wfmanager.rerun_workflow(args["workflow_id"])
    elif args["cmd_object"] == "status":
        if args["workflow_id"]:
            workflow = wfmanager.get_workflow(args["workflow_id"])
            print display_workflow_status(workflow, True)
        else:
            workflows = wfmanager.get_workflows()
            if len(workflows) > 0:
                status = "ID\tNAME\tSTATUS\tELAPSED_TIME\tSTART_TIME\tEND_TIME\n"
                for i, workflow in enumerate(workflows):
                    status += display_workflow_status(workflow)
                    if i<len(workflows)-1: status += "\n"
            else: status = "no workflow available"
            print status
            