#
# Copyright (C) 2012 INRA
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from subprocess import Popen, PIPE

from jflow.component import Component
from jflow.iotypes import OutputFile, OutputFileList, InputFile, InputFileList, Formats

from weaver.function import ShellFunction
from weaver.abstraction import Map
from jflow.abstraction import MultiMap


class BWA (Component):
    
    def define_parameters(self, reference_genome, read1, read2=None, algorithm="aln"):
        self.read1 = InputFileList(read1, Formats.FASTQ)
        self.read2 = None
        if algorithm == "aln":
            self.sai1 = OutputFileList(self.get_outputs('{basename_woext}.sai', self.read1))
        else: 
            self.sai1 = None
        if read2: 
            self.read2 = InputFileList(read2, Formats.FASTQ)
            if algorithm == "aln":
                self.sai2 = OutputFileList(self.get_outputs('{basename_woext}.sai', self.read2))
            else: 
                self.sai2 = None
            self.bam_files = OutputFileList(self.get_outputs('{basename_woext}.bam', [self.read1, self.read2]), Formats.BAM)
        else:
            self.sai2 = None
            self.bam_files = OutputFileList(self.get_outputs('{basename_woext}.bam', self.read1), Formats.BAM)
        self.algorithm = algorithm
        self.reference_genome = InputFile(reference_genome)
        self.stderr = os.path.join(self.output_directory, 'bwa.stderr')
        
    def process(self):        
        if self.algorithm=="bwasw":
            if self.read2:
                bwa = ShellFunction(self.get_exec_path("bwa") + " " + self.algorithm + " " + self.reference_genome + \
                                " $1 $2 2>> " + self.stderr + " | " + self.get_exec_path("samtools") + " view -bS - > $3 2>> " + self.stderr, cmd_format='{EXE} {IN} {OUT}')
                bwasw = MultiMap(bwa, inputs=[self.read1, self.read2], outputs=self.bam_files, includes=self.reference_genome)
            else:
                bwa = ShellFunction(self.get_exec_path("bwa") + " " + self.algorithm + " " + self.reference_genome + \
                                    " $1 2>> " + self.stderr + " | " + self.get_exec_path("samtools") + " view -bS - > $2 2>> " + self.stderr, cmd_format='{EXE} {IN} {OUT}')
                bwasw = Map(bwa, self.read1, self.bam_files, includes=self.reference_genome)
        else:
            reads, sais = [], []
            reads.extend(self.read1)
            sais.extend(self.sai1)
            bwa = ShellFunction(self.get_exec_path("bwa") + " " + self.algorithm + " " + self.reference_genome + \
                                " $1 > $2 2>> " + self.stderr, cmd_format='{EXE} {IN} {OUT}')
            if self.read2:
                reads.extend(self.read2)
                sais.extend(self.sai2)
                bwa_aln = Map(bwa, inputs=reads, outputs=sais, includes=self.reference_genome)
                bwasampe = ShellFunction(self.get_exec_path("bwa") + " sampe " + self.reference_genome + \
                                         " $1 $2 $3 $4 2>> " + self.stderr + " | " + self.get_exec_path("samtools") + " view -bS - > $5 2>> " + self.stderr, cmd_format='{EXE} {IN} {OUT}')
                bwasampe = MultiMap(bwasampe, inputs=[self.sai1, self.sai2, self.read1, self.read2], outputs=self.bam_files, includes=self.reference_genome)
            else:
                bwa_aln = Map(bwa, inputs=reads, outputs=sais, includes=self.reference_genome)
                bwasamse = ShellFunction(self.get_exec_path("bwa") + " samse " + self.reference_genome + \
                                         " $1 $2 2>> " + self.stderr + " | " + self.get_exec_path("samtools") + " view -bS - > $3 2>> " + self.stderr, cmd_format='{EXE} {IN} {OUT}')
                bwasamse = MultiMap(bwasamse, inputs=[self.sai1, self.read1], outputs=self.bam_files, includes=self.reference_genome)

    def get_version(self):
        cmd = [self.get_exec_path("bwa")]
        p = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()
        return stderr.split()[7]
